/*
 * @(#)AsciiPacker.java  1.2  23rd December 2001 Markus Gebhard
 * @(#)AsciiPacker.java  1.1  11th July 2001 Markus Gebhard
 * @(#)AsciiPacker.java  1.0  28th June 2001 Markus Gebhard
 * @(#)AsciiPacker.java  0.1  10th June 2001 Markus Gebhard
 * 
 * This Software is Open Source. You are allowed to change anything,
 * but there are some limitations:
 *
 *  - You must always include the source code
 *  - You are not allowed to remove this Header from the source code
 *  - You have to add a link to http://www.jave.de so people know
 *    where to find the original version and information about it
 *  - You have to let me know about new releases
 *
 *                                      Markus Gebhard  markus@jave.de
 */

public class AsciiPacker{
  
/**  
===============================================================================
Note: This File is the same as the one in used for the Jave (Editor!) Binaries.
To avoid incompatibilities: do not make changes to this file without
informing me:  markus@jave.de
===============================================================================

This class provides some packing/unpacking algorithms for Ascii Images.

If you don't want to write a better compression algorithm and just
use this class for compression/uncompression you only have to look at
those methodes:

    
1) public static String encode(char[][] ch);
      Encodes/compresses 'ch' and returns the compressed String object.

2) public static String encodeOptimized(char[][] ch);
      Encodes/compresses 'ch' as good as possible.
      This methode should be used if there is no need in saving time.

3) public final static char[][] decode(String code);
      Decodes/uncompresses 'code' and returns the field of chars.


There are some aspects about the algorithms in this class:

1) The compressed result hast to be a single line of Ascii characters
   (this is not true when the input contained non Ascii characters)

2) There are several algorithms - the one that seems to fit best will be
   used.

3) Different algorithms will be marked by different characters at the
   beginning of the compressed code
   
4) This marker for the algorithm is followed by the size of the character plate

   E.G. the line:
   
A80 30 ...
   
   has the meaning: Algorithm "A" used for compression, plate size is 80x30,
   "..." is the compressed code.
   So the format of a encoded line is:

ALGORITHM WIDTH HEIGHT CODE

5) If after uncompression there are undefined characters in the plate,
   those characters shall be white spaces.

*/



/**
* Do not let anybody instanciate this class
*/
protected AsciiPacker(){}

/**
* This method is called to encode a field of characters.
* In difference to encode(char[][] ch) this method will take longer
* to execute, but will result in the smallest compressed size possible
* with all the algorithms in this class.
*
* The result is a single line of compressed code.
*/
public static String encodeOptimized(char[][] ch){
  int h=ch.length;
  if (h==0)
    return "B0 0";
  int w=ch[0].length;
  if (w==0)
    return "B0 "+h;

  String code=encodeA(ch);
  String prefix="A";

  //Check the other algorithms
  String codeB=encodeB(ch);
  if (codeB.length()<code.length()){
    prefix="B";
    code=codeB;
  }

  // 
  // FOR EVERY NEW ALGORITHM ADD CHECKS HERE 
  //
  
  return prefix+w+" "+h+" "+code;
}  
  
  
  

/**
* This method is called to encode a field of characters.
* It determins what encoding algorithm to use.
*
* The result is a single line of compressed code.
*/
public static String encode(char[][] ch){
  int h=ch.length;
  if (h==0)
    return "B0 0";
  int w=ch[0].length;
  if (w==0)
    return "B0 "+h;
  
  String code=null;
  String prefix="";
  //Chose the best algorithm
  if (w*h<41){
    prefix="B";
    code=encodeB(ch);
  }else{
    prefix="A";
    code=encodeA(ch);
  }
  
  return prefix+w+" "+h+" "+code;
}  



/**
* Tries to decode the given line of compressed Ascii.
* It will throw a RuntimeException if it is not possible to decode it.
*/
public final static char[][] decode(String code){
  if (code==null || code.length()<4)
    throw new RuntimeException("Unable to decode compressed ASCII:"
                               +"The compressed code doesn not seem to be valid!");  
  
  char algorithm=code.charAt(0);
  switch (algorithm){
    case 'A':
      return decodeA(code);
    case 'B':
      return decodeB(code);
  }
  throw new RuntimeException("Unable to decode compressed ASCII: Algorithm '"
                             +algorithm+"' not known by this program version! "
                             +"Please download a new release.");  
}  

/**
* Decompress code encoded using algorithm A.
*/
protected final static char[][] decodeA(String code){
  int length=code.length();
  
  //1) Algorithm marker  
  int index=1;
//  char algorithm=code.charAt(0);

  //2) Width
  int w=0;
  char ch=code.charAt(index++);
  while(ch>='0' && ch<='9'){
    w*=10;
    w+=ch-'0';
    ch=code.charAt(index++);
  }
  
  //3) Height
  int h=0;
  ch=code.charAt(index++);
  while(ch>='0' && ch<='9'){
    h*=10;
    h+=ch-'0';
    ch=code.charAt(index++);
  }
  
  char[][] chars=new char[h][w];
  int x=0;
  int y=0;  
  while(index<length){
    ch=code.charAt(index++);
    if (ch!='%'){
      chars[y][x++]=ch;
    }else{
      ch=code.charAt(index++);
      if (ch=='%'){
        chars[y][x++]='%';
      }else
      if (ch=='0'){
        //carrige return -> fill the rest of the line with white space
        for (int i=x;i<w;++i)
          chars[y][x++]=' ';
        y+=1;
        x=0;
      }else{
        //how many characters?
        int count=ch-'0';
        ch=code.charAt(index++);
        while(ch>='0' && ch<='9'){
          count*=10;
          count+=ch-'0';
          ch=code.charAt(index++);
        }
        
        if (ch=='%')
          ch=code.charAt(index++);
        
        //count times ch
        for (int i=0;i<count;++i)
          chars[y][x++]=ch;
      }
    }
    
  }

  //Fill up the rest of this line with white space
  while(x<w)
    chars[y][x++]=' ';

  //Fill up the rest of the lines with white space
  for (int yy=y+1;yy<h;++yy)
    for (int xx=0;xx<w;++xx)
      chars[yy][xx]=' ';
  
  return chars;
}  

/**
* Decompress code encoded using algorithm B.
*/
protected final static char[][] decodeB(String code){
  int length=code.length();
  
  //1) Algorithm marker
  int index=1;

  //2) Width
  int w=0;
  char ch=code.charAt(index++);
  while(ch>='0' && ch<='9'){
    w*=10;
    w+=ch-'0';
    ch=code.charAt(index++);
  }
  
  //3) Height
  int h=0;
  ch=code.charAt(index++);
  while(ch>='0' && ch<='9' && index<code.length()){
    h*=10;
    h+=ch-'0';
    ch=code.charAt(index++);
  }
  
  char[][] chars=new char[h][w];

  int i=index;
  int l=code.length();
  for (int y=0;y<h;++y){
    for (int x=0;x<w;++x){
      if (i>=l)
        chars[y][x]=' ';
      else
        chars[y][x]=code.charAt(i);
      ++i;
    }
  }
  return chars;
}  

/**
* Encoding algorithm A:
*       _        
*      / \     _ 
*     / _ \   (_)
*    / ___ \   _ 
*   /_/   \_\ (_)
* rules:
*      %0      : Carrige return
*      %nnx    : nn times character x (x can be anything but a ciffer)
*      %nn%1   : nn times ciffer 1
*      %nn%%   : 1... times character %
*      %%      : the character %
*      x       : the character x (x can be anything but %)
*/
protected static String encodeA(char[][] ch){
  int w=ch[0].length;
  if (w==0)
    return "";
  int h=ch.length;
  if (h==0)
    return "";
  
  
  
  int state=1;
  int x=1;
  int y=0;
  char cc=ch[0][0]; //Current char
  int count=1;      //number of identical subsequent characters
  int eolCount=0;   //number of subsequent CRs
  StringBuffer sc=new StringBuffer(w*h*2+10);
  while (state!=-1){
    switch (state){
      case 1:
        if (y>=h){ //EOT
          if (cc!=' ')
            printA(sc, cc, count);
          state=-1;
          continue;
        }
        if (x>=w){//EOL
          if (cc!=' ')
            printA(sc, cc, count);
          eolCount=1;
          state=2;
          x=0;
          ++y;
          continue;
        }
        if (ch[y][x]==cc){
          ++count;
          ++x;
          continue;
        }else{
          printA(sc, cc, count);
          cc=ch[y][x];
          ++x;
          count=1;
          continue;
        }
      case 2:
        if (y>=h){ //EOT
          state=-1;
          continue;
        }
        if (x>=w){//EOL
          ++eolCount;
          x=0;
          ++y;
          continue;
        }

        if (ch[y][x]==' '){
          count=1;
          cc=ch[y][x];
          ++x;
          state=3;
          continue;
        }else{
          printACR(sc, eolCount);
          cc=ch[y][x];
          ++x;
          count=1;
          state=1;
          continue;
        }
      case 3:
        if (y>=h){ //EOT
          state=-1;
          continue;
        }
        if (x>=w){//EOL
          ++eolCount;
          x=0;
          ++y;
          state=2;
          continue;
        }
        if (ch[y][x]==' '){
          ++count;
          ++x;
          continue;
        }else{
          printACR(sc, eolCount);
          printA(sc, ' ', count);
          cc=ch[y][x];
          ++x;
          count=1;
          state=1;
          continue;
        }
    }
  }
  
  return sc.toString();
}

/** helping methode for encodeA(char[][]) */
protected final static void printACR(StringBuffer sb, int count){
  for (int i=0;i<count;++i)
    sb.append("%0");
}

/** helping methode for encodeA(char[][]) */
protected final static void printA(StringBuffer sb, char ch, int count){
  if (count>2){
    sb.append('%'+String.valueOf(count));
    if (ch>='0' && ch<='9'){
      //ciffer
      sb.append("%"+ch);
    }else
    if (ch=='%'){
      sb.append("%%");
    }else{
      sb.append(ch);
    }
  }else
  if (count==2){ //save space: use xx instead of %2x
    if (ch=='%'){
      sb.append("%%%%");
    }else{
      sb.append(ch);
      sb.append(ch);
    }
  }else{
    //count==1
    if (ch=='%')
      sb.append("%%");
    else
      sb.append(ch);
  }
}

/**
* Encoding algorithm B.
* This algorithm does not really pack the content - it is used
* when compression has more overhead than advantages.
*/
protected static String encodeB(char[][] chars){
  int w=chars[0].length;
  int h=chars.length;
  
  StringBuffer sc=new StringBuffer(w*h*2+10);
  
  //simply append every character
  for (int y=0;y<h;++y)
    for (int x=0;x<w;++x)
      sc.append(chars[y][x]);
     
  //Remove white spaces from the end
  int size=sc.length();
  while (size>0 && sc.charAt(size-1)==' ')
    --size;
  
  return sc.toString().substring(0,size);
}


// Just for stress-test (does not work w/o my own libraries!)  
//  
//  public static void main(String args[]){
//  //  long time0=System.currentTimeMillis();
//    
//    int i=0;
//    
//  char[][] ch1=null;  
//  char[][] ch2=null;
//  String code=null;
//  
//  
//    while(true){
//      try{
//        ch1=getRandomField();
//        code=encodeOptimized(ch1);
//        ch2=decode(code);
//      }catch(Exception e){
//        e.printStackTrace();
//        error(e,ch1,code,ch2);
//      }
//      
//      if (ch1!=null && ch2!=null &&!equals(ch1,ch2)){
//        error("Unequal",ch1,code,ch2);
//      }
//      ++i;
//      if (i%10000==0)
//        System.err.println(i);
//    }
//  //  long time1=System.currentTimeMillis();
//  //  System.err.println((time1-time0)+"ms");
//  }
//  
//  
//  protected static void error(Object o, char[][] ch1, String code, char[][] ch2){
//    System.err.println("=================================");
//    System.err.println("Error: "+o);
//    if (ch1!=null)
//      System.err.println(TextTools.toString(ch1));
//    else  
//      System.err.println("null");
//     
//    System.err.println("=================================");
//    if (ch2!=null)
//      System.err.println(TextTools.toString(ch2));
//    else  
//      System.err.println("null");
//    System.err.println("=================================");
//    System.err.println(code);
//    System.err.println("=================================");
//  }
//  
//  
//  
//  protected static boolean equals(char[][] ch1, char[][] ch2){
//    if (ch1.length!=ch2.length)
//      return false;
//    if (ch1[0].length!=ch2[0].length)
//      return false;
//  
//    for (int y=ch1.length-1;y>=0;--y)
//      for (int x=ch1[0].length-1;x>=0;--x)
//  //  for (int y=0;y<ch1.length;++y)
//  //    for (int x=0;x<ch1[0].length;++x)
//        if (ch1[y][x]!=ch2[y][x])
//          return false;
//  
//    return true;
//  }
//  //
//  
//  protected static char[][] getRandomField(){
//    int w=Toolbox.random(1,20);
//    int h=Toolbox.random(1,20);
//    
//  // System.err.println(w+"x"+h);
//    char[][] ch=new char[h][w];
//    char currentChar=' ';
//    for (int y=0;y<h;++y){
//      for (int x=0;x<w;++x){
//        if (Toolbox.random(0,100)>50){
//          currentChar=(char)Toolbox.random(32,127);
//        
//        
//        
//  //        if (currentChar=='%')
//  //          currentChar='?';
//        }
//        ch[y][x]=currentChar;
//      }
//    }
//        
//  
//    return ch;
//  }
}